#define VERSION 0.2

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "common.h"

/* DOS keyhit / getkeyhit */
#ifdef TARGET_DOS
#include <allegro.h>
#include <pc.h>
#include <conio.h>

#define keysetup() //

#define quitr() (kbhit() ? (getch() == 0x1b ? 1 : 0) : 0)
#define QKEY "ESCape"
#endif

/* Windows keyhit / getkeyhit */
#ifdef TARGET_WIN
#include <windows.h>
#include <conio.h>

#define keysetup() //

#define quitr() (_kbhit() ? (_getch() == 0x1b ? 1 : 0) : 0)
#define QKEY "ESCape"
#endif

/* POSIX keyhit / getkeyhit */
#ifdef TARGET_ALSA
#include <inttypes.h>
#include <math.h>
#include <signal.h>
#include <termios.h>

volatile int quitval = 0;

int uef_baudrate = 1200;
int carrier=0;
float phaseshift=0.0f;

int getkey() {
    int character;
    struct termios orig_term_attr;
    struct termios new_term_attr;

    /* set the terminal to raw mode */
    tcgetattr(fileno(stdin), &orig_term_attr);
    memcpy(&new_term_attr, &orig_term_attr, sizeof(struct termios));
    new_term_attr.c_lflag &= ~(ECHO|ICANON);
    new_term_attr.c_cc[VTIME] = 0;
    new_term_attr.c_cc[VMIN] = 0;
    tcsetattr(fileno(stdin), TCSANOW, &new_term_attr);

    /* read a character from the stdin stream without blocking */
    /*   returns EOF (-1) if no character is available */
    character = fgetc(stdin);

    /* restore the original terminal attributes */
    tcsetattr(fileno(stdin), TCSANOW, &orig_term_attr);

    return character;
}

void handler(int signum)
{
	quitval = 1;
}

void keysetup(void)
{
	if(signal(SIGINT, handler) == SIG_IGN) signal(SIGINT, SIG_IGN);
	if(signal(SIGHUP, handler) == SIG_IGN) signal(SIGHUP, SIG_IGN);
	if(signal(SIGTERM, handler) == SIG_IGN) signal(SIGTERM, SIG_IGN);
}

#define QKEY "CTRL+C"
#define quitr() quitval
#endif

#include "uef.h"
#include "card.h"

#define OUTPUT_HZ 44100
#define BIT_SIZE 36
int bit_size = BIT_SIZE;
//#define BUF_SIZE 16384
#define BUF_SIZE 4096
#define PHASESHIFT 220

//void set_to_wave(signed char *buffer, int numwaves, unsigned char volume)
void set_to_wave(short *buffer, int numwaves, short volume)
{
	int c = BIT_SIZE;
	float halfc;

	halfc = (float)c /2;

	while(c--)
	{
		//buffer[c] = (short)( volume*-sin( ((float)c*3.141592654f*(float)numwaves)/halfc) );
		buffer[c] = (short)( volume*-sin( (((float)c*3.141592654f)*(float)numwaves)/halfc + (phaseshift*3.141592654f)/180) );
		//printf("set_to_wave: buffer[%d] = %d\n",c,buffer[c]);
	}
}

C_UEF infile;
bool (* get_bit_ptr)(void);
int internal_count, internal_count2, internal_target;
unsigned char data;
bool chunk_done, silence;
bool stop_bit;

//int uef_baudrate = 1200;
//int uef_baudrate = 1400;
//int uef_baudrate = 1550;

bool get_bit_implicit_data(void)
{
	bool ret;

	stop_bit = false;

	if(internal_count == 0)
		ret = false;

	if(internal_count == 9) {
		stop_bit = true;
		ret = true;
	}

	if(internal_count > 0 && internal_count < 9)
	{
		ret = ( (data >> (internal_count-1)) &1) ? true : false;
	}

	internal_count++;
	if(internal_count == 10)
	{
		internal_count = 0;

		if(infile.EOC())
			chunk_done = true;
		else
		{
			infile.ReadSome(&data, 1);
		}
	}

	return ret;
}

bool get_bit_data(void)
{
	bool ret;

	ret = ( (data >> (8-internal_count)) &1) ? true : false;

	internal_count2++;
	if(internal_count2 == internal_target)
	{
		chunk_done = true;
	}
	else
	{
		internal_count++;
		if(internal_count == 8)
		{
			internal_count = 0;
			infile.ReadSome(&data, 1);
		}
	}

	return ret;
}

bool get_bit_htone(void)
{
	if(!internal_count)
		chunk_done = true;
	internal_count--;

	return true;
}

bool get_bit_htonedummy(void)
{
	if(internal_count)
	{
		internal_count--;
		return internal_count ? true : false;
	}
	else
	{
		internal_count2--;
		if(!internal_count2)
			chunk_done = true;
	}

	return true;
}

void open_next_chunk(void)
{
	unsigned char low, high;

	chunk_done = true;

	while(chunk_done && !infile.OverRan())
	{
		infile.FindIdMajor(0x1);

		switch(infile.GetId()&255)
		{
			case 0x00 : //implicit data
				get_bit_ptr = get_bit_implicit_data;

				infile.ReadSome(&data, 1);
				internal_count = 0;

				chunk_done = infile.GetLength() ? false : true;
				silence = false;
				stop_bit = false;
			break;

			case 0x02 : //explicit data
				get_bit_ptr = get_bit_data;

				if(infile.GetVersion() >= 9)
				{
					infile.ReadSome(&data, 1);
					internal_target = ((infile.GetLength()-1) << 3) - data;
				}
				else
					internal_target = infile.GetLength() << 3;

				internal_count2 = 0;
				infile.ReadSome(&data, 1);
				internal_count = 0;

				chunk_done = (internal_target>0) ? false : true;
				silence = false;
			break;

			case 0x10 : //high tone
				get_bit_ptr = get_bit_htone;

				infile.ReadSome(&low, 1);
				infile.ReadSome(&high, 1);
				internal_count = low | (high << 8);

				if(infile.GetVersion() < 9) {
					internal_count = (uef_baudrate*internal_count) / 1000;
					if (carrier > 0) {
						internal_count = carrier;
					}
                                        //printf("high tone %ld\n",internal_count);
				}

				chunk_done = internal_count ? false : true;
				silence = false;
			break;

			case 0x11 : //high tone with dummy
				get_bit_ptr = get_bit_htonedummy;

				infile.ReadSome(&low, 1);
				infile.ReadSome(&high, 1);
				internal_count = (low | (high << 8)) + 1;
				infile.ReadSome(&low, 1);
				infile.ReadSome(&high, 1);
				internal_count2 = low | (high << 8);

				if(infile.GetVersion() < 9)
				{
					internal_count = (uef_baudrate*internal_count) / 1000;
					internal_count2 = (uef_baudrate*internal_count) / 1000;
				}

				/* time for extra byte */
				internal_count += 10;

				chunk_done = false;
				silence = true;
			break;

			case 0x12 :	//silence
				get_bit_ptr = get_bit_htone;

				infile.ReadSome(&low, 1);
				infile.ReadSome(&high, 1);
				internal_count = low | (high << 8);

				if(infile.GetVersion() < 9)
					internal_count = (uef_baudrate*internal_count) / 1000;

				chunk_done = internal_count ? false : true;
				silence = true;
			break;

			case 0x13 :	//change of baud rate
				infile.ReadSome(&low, 1);
				infile.ReadSome(&high, 1);
				uef_baudrate = low | (high << 8);
			break;
		}
	}

	if(infile.OverRan())
		silence = true;
}

bool get_bit(void)
{
	while(chunk_done && !infile.OverRan())
		open_next_chunk();

	return get_bit_ptr();
}

bool (* begin_sound)(int freq, int bufsize, char *devicename);
void (* end_sound)(void);
//unsigned char *(* get_next_pointer)(void);
short *(* get_next_pointer)(void);
void (* done_with_pointer)(void);
void (* pause_sound)(void);
void (* resume_sound)(void);

int main(int argc, char **argv)
{
        extern char *optarg;
        extern int optind;
        int c, err = 0;

        //signed char one[BIT_SIZE], zero[BIT_SIZE], *bitbuf;
        short one[BIT_SIZE], zero[BIT_SIZE], *bitbuf;
        bool quit = false;
        //unsigned char *sndbuf;
        short *sndbuf;
        int bitpos = BIT_SIZE << 16, bufpos;
        int key;
        char *alsadevice = NULL;
	char *ueffilename = NULL;

        char *baudratearg = NULL;
        char *carrierarg = NULL;
        char *phaseshiftarg = NULL;
        static char usage[] = "FreeUEF-ALSA version %.1f\n\nusage: %s [-b baudrate] [-c carrier_length] [-p phaseshift] UEF_file [ALSA_device]\n\nPlay <UEF filename> using ALSA. Default sound device is plughw:1,0, but\nyou can specify a different sound device as the 2nd arg. For example:\n\n  freeuef-alsa game.uef plughw:0,0\n\nTurbo Mode\nYou can get load times a little faster by playing with the baudrate, the carrier length (or 'high tone') and also the phaseshift of the sinewaves. Have a look at tools like PlayUEF and various discussions on stardot.org.uk to see what is possible.\n\nExamples\n\n%s -b 1550 -c 60 -p 220 game.uef\n\n%s -b 1400 game.uef\n";


        while ((c = getopt(argc, argv, "b:c:p:")) != -1) {
                switch (c) {
                case 'b':
                        baudratearg = optarg;
			uef_baudrate = atoi(baudratearg);
                        break;
                case 'c':
                        carrierarg = optarg;
			carrier = atoi(carrierarg);
                        break;
		case 'p':
			phaseshiftarg = optarg;
			phaseshift = atof(phaseshiftarg);
			break;
                case '?':
                        err = 1;
                        break;
                }
        }
        if ((optind+1) > argc) {
                /* need at least one argument (change +1 to +2 for two, etc. as needeed) */

                fprintf(stderr, usage, VERSION, argv[0],argv[0],argv[0]);
                exit(1);
        } else if (err) {
                fprintf(stderr, usage, VERSION,argv[0],argv[0],argv[0]);
                exit(1);
        }
        /* see what we have */
        //printf("baudrate = %s\n", baudratearg);
        //printf("carrier = %s\n", carrierarg);
        //printf("phaseshift = %s, %.0f\n", phaseshiftarg,phaseshift);

        if (optind < argc) {
        	/* these are the arguments after the command-line options */
		ueffilename = argv[optind];
		//printf("ueffilename = %s\n", ueffilename);
		optind++;
		if (optind <argc) {
			alsadevice = argv[optind];
			printf("alsadevice = %s\n",alsadevice);
		}
        } else {
                printf("no arguments left to process\n");
        }

	printf("FreeUEF-ALSA %.2f with optional turbo hacks\n",VERSION);
        if (uef_baudrate != 1200) {
		printf("   set baudrate to %d\n",uef_baudrate);
	}
	if (carrier > 0) {
		printf ("   force carrier high tone length to %d cycles\n",carrier);
	}
	if (phaseshift > 0) {
		printf ("   phase shift to %.0f degrees\n",phaseshift);
	}

	keysetup();

	//Open(char *name, char *mode, unsigned __int8 minor, unsigned __int8 major) 
	if(!infile.Open(ueffilename, "r", 9, 0))
	{
		begin_sound = begin_sound_card;
		end_sound = end_sound_card;
		get_next_pointer = get_next_pointer_card;
		done_with_pointer = done_with_pointer_card;
		pause_sound = pause_sound_card;
		resume_sound = resume_sound_card;

		open_next_chunk();
		// 2nd arg to set_to_wave is 'short' volume magnitude
		//  For me (kernelcrash) a one with 40000, and a 
		//    zero with 56000 work with ALSA master and PCM vol set
		//    to full 100%

		//set_to_wave(one, 2, 90);
		set_to_wave(one, 2, 40000);

		//set_to_wave(zero, 1, 120);
		set_to_wave(zero, 1, 56000);

		if(!begin_sound(OUTPUT_HZ, BUF_SIZE*2,alsadevice))
		{
			printf("ERROR : cannot open sound device\n");
			exit(3);
		}

		printf("Press ENTER when ready - %s to quit at any time (may take a while)\n", QKEY);
		getchar();
		printf("\nPress SPACE to pause playback (generally when your Electron clicks it relay)\n");

		int bufsize = get_buff_size();

		printf("\n");
		int block=0;
		while(!infile.OverRan() && !quit)
		{
			printf("\rblock %d     ",block++);
			if(quitr())
			{
				quit = true;
				break;
			}

			sndbuf = get_next_pointer();
			//printf("sndbuf = 0x%" PRIXPTR "\n", (uintptr_t)sndbuf);
			for(bufpos = 0; bufpos < (bufsize/4); bufpos++)
			{
				//printf("bufpos = %d\n",bufpos);
				//if(bitpos >= (BIT_SIZE << 16))
				if(bitpos >= (BIT_SIZE << 16))
				{
					bitbuf = get_bit() ? one : zero;
					// 20180206 Leaving stop bit as same length, as cannot get it to work
					if (stop_bit) {
						bit_size = BIT_SIZE;
					} else {
						bit_size = BIT_SIZE;
					}
					//printf("internal_count = %d\n",internal_count);
					bitpos -= BIT_SIZE << 16;
				}

				//sndbuf[bufpos*2] = 127 + (silence ? 0 : bitbuf[bitpos >> 16]);
				sndbuf[bufpos*2] = 0 + (silence ? 0 : bitbuf[bitpos >> 16]);
				//sndbuf[bufpos*2+1] = 127 + (silence ? 0 : bitbuf[bitpos >> 16]);
				sndbuf[bufpos*2+1] = 0 + (silence ? 0 : bitbuf[bitpos >> 16]);
				bitpos += uef_baudrate*((BIT_SIZE*65536) / OUTPUT_HZ);
			}
			done_with_pointer();
			key = getkey();
			if (key == 0x20) {
				printf("\nPaused. Press ENTER to continue (generally when you see 'Searching' on the Electron)\n");
				pause_sound();
				read(1,&key,1);
				printf("UnPaused ...\n");
				resume_sound();
			}

		}
		
		end_sound();

		if(infile.OverRan())
			printf("Done\n");
		else
			printf("Aborted\n");
	}
	else
	{
		printf("ERROR : cannot open UEF\n");
		exit(2);
	}

	return 0;
}
